<?php

    include '../../../common/include/common.php';
    include_once '../../config.php';
    $module = modules::GATK;
    printHeader($module, "Picard | Tool Documentation Index", "Guide");
?>

<div class='row-fluid'>

<?php printGATKDocsNav($module); ?>

<div class='span9'>

<link type='text/css' rel='stylesheet' href='picardDoc.css'>

<#include "common.html"/>

<#macro emitGroup group>
    <div class="accordion-group">
        <div class="accordion-heading">
            <a class="accordion-toggle" data-toggle="collapse" data-parent="#index" href="#${group.id}">
                <h4>${group.name}</h4>
            </a>
        </div>
        <div class="accordion-body collapse" id="${group.id}">
            <div class="accordion-inner">
                <p class="lead">${group.summary}</p>
                <table class="table table-striped table-bordered table-condensed">
                    <tr>
                        <th>Name</th>
                        <th>Summary</th>
                    </tr>
                    <#list data as datum>
                        <#if datum.group == group.name>
                            <tr>
                                <td><a href="${datum.filename}">${datum.name}</a></td>
                                <td>${datum.summary}</td>
                            </tr>
                        </#if>
                    </#list>
                </table>
            </div>
        </div>
    </div>
</#macro>

<h1 id="top">Picard Tool Documentation Index
    <small>${version}</small>
</h1>
<div class="accordion" id="index">
    <#assign seq = ["engine", "tools", "other", "utilities"]>
	<#list seq as supercat>
		<br />
		<#list groups?sort_by("name") as group>
			<#if group.supercat == supercat>
				<@emitGroup group=group/>
			</#if>
		</#list>
	</#list>
</div>

<@footerInfo />
<@footerClose />

</div></div>

<?php

    printFooter($module);

?>