<?php

    include '../../../common/include/common.php';
    include_once '../../config.php';
    $module = modules::GATK;
    printHeader($module, "Picard | Tool Documentation Index", "Guide");
?>

<link type='text/css' rel='stylesheet' href='picardDoc.css'>

<div class='row-fluid' id="top">

	<#include "common.html"/>

	<?php $group = '${group}'; ?>

	<section class="span4">
		<aside class="well">
			<a href="index.html"><h4><i class='fa fa-chevron-left'></i> Back to Tool Docs Index</h4></a>
		</aside>
		<aside class="well">
			<h2>Categories</h2>
			<@getCategories groups=groups />
		</aside>
		<?php getForumPosts( '${name}' ) ?>

	</section>

	<div class="span8">

		<h1>${name}</h1>

		<p class="lead">${summary}</p>

		<#if group?? >
			<h3>Category
				<small> ${group}</small>
			</h3>
		</#if>
		<hr>
		<h2>Overview</h2>
		${description}

			<#-- Create the metrics summary -->
			<#if metrics?size != 0>
				<p>This table summarizes the values that are specific to this metric.</p>
				<table class="table table-striped table-bordered table-condensed">
					<thead>
					<tr>
						<th>Metric</th>
						<th>Summary</th>
					</tr>
					</thead>
					<#list metrics as metric>
					<tr>
						<td>${metric.name}<br />
						<td>${metric.summary}</td>
					</tr>
					</#list>
					</tbody>
				</table>
			</#if>

			<@footerInfo />
			<@footerClose />

	</div>

	<?php printFooter($module); ?>