body {
    margin: 40px;
    font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
    font-size: 13px;
    line-height: 18px;
    color: #333333;
    background-color: #ffffff;
}
a {
    color: #269abc;
    text-decoration: none;
}
a:hover {
    color: #004d8b;
    text-decoration: underline;
}
.label,
.badge {
    font-size: 11px;
    font-weight: bold;
    line-height: 14px;
    color: #ffffff;
    vertical-align: baseline;
    white-space: nowrap;
    text-shadow: 0 -1px 0 rgba(0, 0, 0, 0.25);
    background-color: #999999;
}
.label {
    padding: 1px 4px 2px;
    -webkit-border-radius: 3px;
    -moz-border-radius: 3px;
    border-radius: 3px;
}
.badge {
    padding: 1px 9px 2px;
    -webkit-border-radius: 9px;
    -moz-border-radius: 9px;
    border-radius: 9px;
}
a.label:hover,
a.badge:hover {
    color: #ffffff;
    text-decoration: none;
    cursor: pointer;
}
.label-important,
.badge-important {
    background-color: #b94a48;
}
.label-important[href],
.badge-important[href] {
    background-color: #953b39;
}
.label-warning,
.badge-warning {
    background-color: #e44a00;
}
.label-warning[href],
.badge-warning[href] {
    background-color: #b13900;
}
.label-success,
.badge-success {
    background-color: #468847;
}
.label-success[href],
.badge-success[href] {
    background-color: #356635;
}
.label-info,
.badge-info {
    background-color: #3a87ad;
}
.label-info[href],
.badge-info[href] {
    background-color: #2d6987;
}
.label-inverse,
.badge-inverse {
    background-color: #333333;
}
.label-inverse[href],
.badge-inverse[href] {
    background-color: #1a1a1a;
}
table {
    max-width: 100%;
    background-color: transparent;
    border-collapse: collapse;
    border-spacing: 0;
}
.table {
    width: 100%;
    margin-bottom: 40px;
}
.table th,
.table td {
    padding: 8px;
    line-height: 18px;
    text-align: left;
    vertical-align: top;
    border-top: 1px solid #dddddd;
}
.table th {
    font-weight: bold;
}
.table thead th {
    vertical-align: bottom;
}
.table tbody + tbody {
    border-top: 2px solid #dddddd;
}
.table-condensed th,
.table-condensed td {
    padding: 4px 5px;
}
.table-bordered {
    border: 1px solid #dddddd;
    border-collapse: separate;
    *border-collapse: collapsed;
    border-left: 0;
    -webkit-border-radius: 4px;
    -moz-border-radius: 4px;
    border-radius: 4px;
}
.table-bordered th,
.table-bordered td {
    border-left: 1px solid #dddddd;
}
.table-striped tbody tr:nth-child(odd) td,
.table-striped tbody tr:nth-child(odd) th {
    background-color: #f9f9f9;
}
.table tbody tr:hover td,
.table tbody tr:hover th {
    background-color: #f5f5f5;
}
pre code {
    padding: 0;
    color: #F5F5F5;
    background-color: transparent;
    border: 0;
}
.accordion-heading {
    font-size: 24px;
    line-height: 30px;
}
.lead {
    font-size: 18px;
}
.hide_me_html {
    display: none;
}
.well {
    font-size: 14px;
    padding-bottom: 20px;
}
.cozy {
    font-size: 12px;
}