package picard.sam.util;

/**
 * Little struct-like class to hold a record index, the index of the corresponding representative read, and duplicate set size information.
 */

public class RepresentativeReadIndexer {
    public int readIndexInFile = -1;
    public int setSize = -1;
    public int representativeReadIndexInFile = -1;
}
