package picard.fingerprint;
/*
 * The MIT License
 *
 * Copyright (c) 2020 The Broad Institute
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

import picard.util.MathUtil;

public class CappedHaplotypeProbabilities extends HaplotypeProbabilitiesUsingLogLikelihoods {
    private final double cap;

    // cap should be negative to indicate that you can never be too sure of anything (since the log likelihood is the
    //    log probability of an error, it's negative)
    public CappedHaplotypeProbabilities(final HaplotypeProbabilities haplotypeProbabilities, double cap) {
        super(haplotypeProbabilities.getHaplotype());
        this.cap = cap;
        final double[] logLikelihoods = haplotypeProbabilities.getLogLikelihoods();
        final double[] cappedLogLikelihoods = MathUtil.subtractMax(logLikelihoods);
        this.setLogLikelihoods(MathUtil.capFromBelow(cappedLogLikelihoods, cap));
    }


    @Override
    public HaplotypeProbabilities deepCopy() {
        return new CappedHaplotypeProbabilities(this, cap);
    }
}
